import argparse
import os

from .lib import search


def cli():
    parser = argparse.ArgumentParser(
        description="Search for abandoned and deprecated python packages"
    )
    parser.add_argument(
        "path",
        type=str,
        help="Path to a virtualenv to search",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase output verbosity",
    )
    parser.add_argument(
        "--format",
        choices=["text", "json"],
        default="text",
        help="Output format",
    )
    args = parser.parse_args()

    gh_token = os.environ.get("GH_TOKEN")
    if not (gh_token):
        raise Exception("GH_TOKEN environment variable must be set")

    return search(gh_token, args.path, args.verbose, args.format)



import json
import sys
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from unittest.mock import patch

import pytest
import responses
from rich.console import Console

from pip_abandoned import lib

if sys.version_info < (3, 10):
    from importlib_metadata import Distribution
else:
    from importlib.metadata import Distribution

# Disable Rich formatting so we can more easily make assertions about text output
lib.console = Console(force_terminal=True, _environ={"TERM": "dumb"}, soft_wrap=True)


def get_dist_fixture(name):
    return Distribution.at(Path(".") / "tests" / "fixture_data" / name)


@pytest.fixture
def mock_distributions_no_packages():
    with patch("pip_abandoned.lib.distributions") as mock:
        mock.return_value = []
        yield mock


@pytest.fixture
def mock_distributions_inactive():
    with patch("pip_abandoned.lib.distributions") as mock:
        mock.return_value = [get_dist_fixture("inactive-1.0.0.dist-info")]
        yield mock


@pytest.fixture
def mock_distributions_readme():
    with patch("pip_abandoned.lib.distributions") as mock:
        mock.return_value = [get_dist_fixture("readme-1.0.0.dist-info")]
        yield mock


@pytest.fixture
def mock_distributions_homepage():
    with patch("pip_abandoned.lib.distributions") as mock:
        mock.return_value = [get_dist_fixture("home-page-1.0.0.dist-info")]
        yield mock


@pytest.fixture
def mock_distributions_project_urls():
    with patch("pip_abandoned.lib.distributions") as mock:
        mock.return_value = [get_dist_fixture("project-urls-1.0.0.dist-info")]
        yield mock


@pytest.fixture
def mock_all_errors():
    with patch("pip_abandoned.lib.distributions") as mock:
        mock.return_value = [
            get_dist_fixture("inactive-1.0.0.dist-info"),
            get_dist_fixture("home-page-1.0.0.dist-info"),
            get_dist_fixture("readme-1.0.0.dist-info"),
        ]
        yield mock


class TestSearch:
    def test_no_packages(self, mock_distributions_no_packages):
        with pytest.raises(Exception) as exc:
            lib.search("fake_token", "/fake/path", 0)
        assert "Couldn't find any packages in /fake/path" in str(exc)

    def test_inactive(self, mock_distributions_inactive):
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "Packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "No packages with a [maintained|no] badge were found" in stdout
        assert "No packages associated with archived GitHub repos were found" in stdout
        assert exit_code == 1

    def test_unmaintained_badge(self, mock_distributions_readme):
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "No packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "Packages with a [maintained|no] badge were found" in stdout
        assert "No packages associated with archived GitHub repos were found" in stdout
        assert exit_code == 1

    @responses.activate
    def test_home_page_archived(self, mock_distributions_homepage):
        responses.add(
            responses.POST,
            "https://api.github.com/graphql",
            json={"data": {"home_page": {"isArchived": True}}},
            status=200,
        )
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "No packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "No packages with a [maintained|no] badge were found" in stdout
        assert "Packages associated with archived GitHub repos were found:" in stdout
        assert exit_code == 1

    @responses.activate
    def test_home_page_not_archived(self, mock_distributions_homepage):
        responses.add(
            responses.POST,
            "https://api.github.com/graphql",
            json={"data": {"home_page": {"isArchived": False}}},
            status=200,
        )
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "No packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "No packages with a [maintained|no] badge were found" in stdout
        assert "No packages associated with archived GitHub repos were found" in stdout
        assert exit_code == 0

    @responses.activate
    def test_project_urls_archived(self, mock_distributions_project_urls):
        responses.add(
            responses.POST,
            "https://api.github.com/graphql",
            json={"data": {"project_urls": {"isArchived": True}}},
            status=200,
        )
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "No packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "No packages with a [maintained|no] badge were found" in stdout
        assert "Packages associated with archived GitHub repos were found:" in stdout
        assert exit_code == 1

    @responses.activate
    def test_project_urls_not_archived(self, mock_distributions_project_urls):
        responses.add(
            responses.POST,
            "https://api.github.com/graphql",
            json={"data": {"project_urls_": {"isArchived": False}}},
            status=200,
        )
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "No packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "No packages with a [maintained|no] badge were found" in stdout
        assert "No packages associated with archived GitHub repos were found" in stdout
        assert exit_code == 0

    @responses.activate
    def test_all_errors(self, mock_all_errors):
        responses.add(
            responses.POST,
            "https://api.github.com/graphql",
            json={"data": {"home_page": {"isArchived": True}}},
            status=200,
        )
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0)
            stdout = buf.getvalue()

        assert (
            "Packages with the trove classifier 'Development Status :: 7 - Inactive' were found"
            in stdout
        )
        assert "Packages with a [maintained|no] badge were found" in stdout
        assert "Packages associated with archived GitHub repos were found:" in stdout
        assert exit_code == 1

    @responses.activate
    def test_json_output(self, mock_all_errors):
        responses.add(
            responses.POST,
            "https://api.github.com/graphql",
            json={"data": {"home_page": {"isArchived": True}}},
            status=200,
        )
        with StringIO() as buf, redirect_stdout(buf):
            exit_code = lib.search("fake_token", "/fake/path", 0, "json")
            stdout = buf.getvalue()

        assert json.loads(stdout) == {
            "inactive": ["inactive"],
            "unmaintained": ["readme"],
            "archived": ["home-page"],
        }
        assert exit_code == 1


class TestGetGitHubRepo:
    def test_no_matches(self):
        dist = get_dist_fixture("inactive-1.0.0.dist-info")
        assert lib.get_github_repo_url(dist) is None

    def test_home_page_match(self):
        dist = get_dist_fixture("home-page-1.0.0.dist-info")
        expected = "https://github.com/chris48s/does-not-exist"
        assert lib.get_github_repo_url(dist) == expected

    def test_project_urls_one_match(self):
        dist = get_dist_fixture("project-urls-1.0.0.dist-info")
        expected = "https://github.com/chris48s/does-not-exist"
        assert lib.get_github_repo_url(dist) == expected

    def test_project_urls_multiple_matches(self):
        dist = get_dist_fixture("multiple-matches-1.0.0.dist-info")
        assert lib.get_github_repo_url(dist) is None


class TestGitHubRepoOrNull:
    @pytest.mark.parametrize(
        "url",
        [
            "https://github.com/chris48s/does-not-exist",
            "https://github.com/chris48s/does-not-exist/",
            "https://github.com/chris48s/does-not-exist.git",
            "https://github.com/chris48s/does-not-exist#readme",
            "https://github.com/chris48s/does-not-exist/#readme",
        ],
    )
    def test_valid(self, url):
        assert (
            lib.github_repo_url_or_none(url)
            == "https://github.com/chris48s/does-not-exist"
        )

    @pytest.mark.parametrize(
        "url",
        [
            "https://example.com",
            "https://github.com/chris48s/does-not-exist/blob/main/README.md",
            "https://chris48s.github.io/does-not-exist",
        ],
    )
    def test_invalid(self, url):
        assert lib.github_repo_url_or_none(url) is None



import json
import logging
import sys
from urllib.parse import urlparse, urlunparse

import requests
from rich import print_json
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

if sys.version_info < (3, 10):
    from importlib_metadata import Prepared, distributions
else:
    from importlib.metadata import Prepared, distributions

logging.basicConfig(
    format="%(message)s",
    handlers=[RichHandler(show_time=False, console=Console(stderr=True))],
)
logger = logging.getLogger(__name__)

console = Console()


def set_log_level(verbosity):
    if verbosity == 0:
        logger.setLevel(logging.ERROR)
    elif verbosity == 1:
        logger.setLevel(logging.WARNING)
    elif verbosity == 2:
        logger.setLevel(logging.INFO)
    elif verbosity >= 3:
        logger.setLevel(logging.DEBUG)


def github_repo_url_or_none(url):
    if url:
        parsed_url = urlparse(url)
        path_parts = [part for part in parsed_url.path.split("/") if part]
        if parsed_url.netloc == "github.com" and len(path_parts) == 2:
            return strip_suffixes(
                urlunparse(
                    (
                        parsed_url.scheme,
                        parsed_url.netloc,
                        parsed_url.path,
                        parsed_url.params,
                        parsed_url.query,
                        "",  # strip fragment (e.g: #readme) if present
                    )
                )
            )
    return None


def strip_suffixes(url):
    if url.endswith(".git"):
        return url[:-4]
    if url.endswith("/"):
        return url[:-1]
    return url


def get_github_repo_url(distribution):
    urls = set()

    if home_page := github_repo_url_or_none(distribution.metadata.get("Home-page")):
        urls.add(home_page)

    if distribution.metadata.get_all("Project-URL"):
        for classifier in distribution.metadata.get_all("Project-URL"):
            try:
                _, url = classifier.split(", ")
                if project_url := github_repo_url_or_none(url):
                    urls.add(project_url)
            except ValueError:
                pass

    urls = list(urls)

    if len(urls) > 1:
        logger.warning(
            f"Found multiple candidate GitHub repo URLs for package {distribution.metadata.get('name')}: {', '.join(urls)}. Skipping"
        )
    if len(urls) == 1:
        return urls[0]
    return None


def get_graphql_query(dist_urls):
    query = "query {\n"
    for dist, repo in dist_urls:
        owner, name = [part for part in urlparse(repo).path.split("/") if part]
        slug = Prepared.normalize(dist.name)
        query += (
            f'  {slug}: repository(owner: "{owner}", name: "{name}") {{ isArchived }}\n'
        )
    query += "}"
    return query


def query_github_api(gh_token, query):
    logger.info(f"Querying GitHub API:\n{query}")

    resp = requests.post(
        "https://api.github.com/graphql",
        json={"query": query},
        headers={"Authorization": f"token {gh_token}"},
    )
    resp.raise_for_status()
    body = resp.json()
    logger.info(f"Response from GitHub API:\n{json.dumps(body, indent=2)}")

    if body.get("errors"):
        logger.warning(
            f"Encountered errors calling GitHub API:\n{json.dumps(body['errors'], indent=2)}"
        )

    return body["data"]


def get_archived_packages(dist_urls, api_data):
    archived_packages_normalized_names = [
        k for k, v in api_data.items() if v and v.get("isArchived")
    ]

    return [
        (dist, repo)
        for dist, repo in dist_urls
        if Prepared.normalize(dist.name) in archived_packages_normalized_names
    ]


def is_inactive(distribution):
    classifiers = distribution.metadata.get_all("Classifier", [])
    return "Development Status :: 7 - Inactive" in classifiers


def has_maintained_no_badge(distribution):
    description = distribution.metadata.get("Description", "")
    return "//img.shields.io/maintenance/no" in description


def output_package_repo_table(packages):
    table = Table(show_header=True)

    table.add_column("Package")
    table.add_column("Repo")

    for package, repo in packages:
        table.add_row(package.name, repo)

    console.print(table)


def output_package_table(packages):
    table = Table(show_header=True)

    table.add_column("Package")

    for package in packages:
        table.add_row(package.name)

    console.print(table)


def output_console(inactive, unmaintained, archived):
    console.print("\n")
    if len(inactive) == 0:
        console.print(
            "[green]✔[/] No packages with the trove classifier [bold white]'Development Status :: 7 - Inactive'[/] were found"
        )
    else:
        console.print(
            "[red]✖[/] Packages with the trove classifier [bold white]'Development Status :: 7 - Inactive'[/] were found:"
        )
        output_package_table(inactive)
    console.print("\n")

    if len(unmaintained) == 0:
        console.print(
            "[green]✔[/] No packages with a [white on black bold]\x5Bmaintained[/]|[white on red bold]no][/] badge were found"
        )
    else:
        console.print(
            "[red]✖[/] Packages with a [white on black bold]\x5Bmaintained[/]|[white on red bold]no][/] badge were found:"
        )
        output_package_table(unmaintained)
    console.print("\n")

    if len(archived) == 0:
        console.print(
            "[green]✔[/] No packages associated with archived GitHub repos were found"
        )
    else:
        console.print(
            "[red]✖[/] Packages associated with archived GitHub repos were found:"
        )
        output_package_repo_table(archived)
    console.print("\n")


def output_json(inactive, unmaintained, archived):
    print_json(
        data={
            "inactive": [p.name for p in inactive],
            "unmaintained": [p.name for p in unmaintained],
            "archived": [p.name for p, _ in archived],
        }
    )


def search(gh_token, path, verbosity, format_="text"):
    set_log_level(verbosity)

    dists = list(distributions(path=[path]))
    if len(dists) == 0:
        raise Exception(f"Couldn't find any packages in {path}")

    inactive_packages = [dist for dist in dists if is_inactive(dist)]
    unmaintained_packages = [dist for dist in dists if has_maintained_no_badge(dist)]

    dist_urls = []
    for distribution in dists:
        url = get_github_repo_url(distribution)
        if url:
            dist_urls.append((distribution, url))

    archived_packages = []
    if len(dist_urls) > 0:
        query = get_graphql_query(dist_urls)
        archived_packages = get_archived_packages(
            dist_urls, query_github_api(gh_token, query)
        )

    if format_ == "json":
        output_json(inactive_packages, unmaintained_packages, archived_packages)
    else:
        output_console(inactive_packages, unmaintained_packages, archived_packages)

    if (
        len(inactive_packages) == 0
        and len(unmaintained_packages) == 0
        and len(archived_packages) == 0
    ):
        return 0
    return 1
